@@ -16,8 +16,8 @@ class Normalizer():
             logging.root.setLevel(logging.DEBUG)
         self.filename = filename
         self.tokenizer_name = tokenizer_name
-        self.normalizer_result = {'original': '', 'normalized': '', 'map': []}
-        self.data = dict()
+        self.normalizer_result = {'original': '', 'normalized': '', 'map': [], 'r_map': []}
+        self.content = dict()
 
     @property
     def name(self):
@@ -31,6 +31,14 @@ class Normalizer():
     def result(self):
         return self.normalizer_result
 
+    @property
+    def data(self):
+        return self.content
+    
+    @data.setter
+    def data(self, obj):
+        self.content = obj
+
     def update_str_with_chmap(self, value, chmap):
         """This function zooms through a string *value*, replaces characters
         according to *chmap* dictionary object, and returns the updated string.
@@ -118,7 +126,7 @@ class Normalizer():
                         subtrie[actions[action][parameter_keylet]] = parameter_value
                 else:
                     subtrie[actions[action][parameter_key]] = parameter_value
-        self.data = trie
+        self.content = trie
         return True
 
     def chargroup(self, s):
@@ -134,6 +142,29 @@ class Normalizer():
             return 2
         return 0
 
+    def reverse_map(self, m):
+        """This function takes character location map in a form it is stored at self.normalizer_result['map']
+        and returns list where item index is character index in original string, and item value is pair
+        of lowest and highest character index in normalized string.
+        """
+        ret = [None] * (m[-1] + 1)
+        for i in range(len(m)):
+            if ret[m[i]] is None:
+                ret[m[i]] = [i, i]
+            else:
+                ret[m[i]][1] = i
+        j = 0
+        for i in range(0, len(ret)):
+            if ret[i] is not None:
+                j = i
+                for k in range(0, j):
+                    ret[k] = ret[j]
+                break
+        for i in range(j+1, len(ret)):
+            if ret[i] is None:
+                ret[i] = ret[i-1]
+        return ret
+
     def normalize(self, source_string, word_separator=' ', normalizer_option=0):
         """This function zooms through the provided string character by character
         and returns string which is normalized representation of a given string.
@@ -145,14 +176,15 @@ class Normalizer():
         """
         assert len(word_separator) == 1, 'word_separator must be single character'
         # TODO: review for refactoring
-        self.normalizer_result = {'original': source_string, 'normalized': '', 'map': []}
+        self.normalizer_result = {'original': source_string, 'normalized': '', 'map': [], 'r_map': []}
         if source_string == '':
             return ''
         original_string = source_string
-        subtrie = self.data
+        subtrie = self.content
         if '_settings' in subtrie and 'bypass' in subtrie['_settings'] and subtrie['_settings']['bypass'] == '1':
             self.normalizer_result['normalized'] = original_string
             self.normalizer_result['map'] = [i for i in range(len(original_string))]
+            self.normalizer_result['r_map'] = [(i, i) for i in range(len(original_string))]
             return original_string
         if '_settings' not in subtrie or 'cs' not in subtrie['_settings'] or subtrie['_settings']['cs'] != '1':
             original_string = original_string.lower()
@@ -177,8 +209,8 @@ class Normalizer():
         added_separator = False
         while current_index < total_length:
             character = original_string[current_index]
-            if character in self.data['_chmap']:
-                character = self.data['_chmap'][character]
+            if character in self.content['_chmap']:
+                character = self.content['_chmap'][character]
             this_group = self.chargroup(character)
             on_the_right = False
             added_separator = False
@@ -193,7 +225,7 @@ class Normalizer():
                     began_reading = False
                     on_the_right = True
                     added_separator = True
-            if not (subtrie is self.data) and character in self.data and temp_index == -1:
+            if not (subtrie is self.content) and character in self.content and temp_index == -1:
                 # mark this as potential head
                 temp_index, temp_buffer, t_map = current_index, buffer, list(b_map)
             if character in subtrie:
@@ -226,7 +258,7 @@ class Normalizer():
                     last_buffer = buffer
                     last_replacement = subtrie['~_']
                     l_map = [b_map[0] for i in range(len(last_replacement))]
-                if '~_' in subtrie and on_the_left and on_the_right:
+                if '~_' in subtrie and ((on_the_left or on_the_right) or '~m' in subtrie):
                     # now buffer has token to be replaced
                     buffer = subtrie['~_'] + word_separator #if not buffer.endswith(word_separator) else ''
                     b_map = [b_map[0] for i in range(len(buffer))]
@@ -261,7 +293,7 @@ class Normalizer():
                         buffer = word_separator + buffer
                         b_map.insert(0, current_index)
                     temp_index = -1
-                subtrie = self.data
+                subtrie = self.content
                 if temp_index > -1:
                     current_index, buffer, b_map = temp_index, temp_buffer, list(t_map) # plain jumping back which causes performance hit, think about better solution
                     temp_index, temp_buffer, t_map = -1, '', []
@@ -280,10 +312,10 @@ class Normalizer():
                 buffer = character
                 b_map = [current_index for x in character]
                 on_the_left = False
-                if character in self.data:
+                if character in self.content:
                     on_the_left = added_separator or last_character == word_separator
                     began_reading = True
-                    subtrie = self.data[character]
+                    subtrie = self.content[character]
             last_group = this_group
             last_character = character
             current_index += 1
@@ -291,7 +323,7 @@ class Normalizer():
         # DRY!
         on_the_right = True
         on_the_left = this_fragment == '' or this_fragment[-1:] == word_separator
-        if '~_' in subtrie and on_the_left and on_the_right:
+        if '~_' in subtrie and ((on_the_left or on_the_right) or '~m' in subtrie):
             # now buffer has token to be replaced
             buffer = subtrie['~_'] + word_separator #if not buffer.endswith(word_separator) else ''
             b_map = [b_map[0] for i in range(len(buffer))]
@@ -324,9 +356,10 @@ class Normalizer():
             normalized = word_separator.join(sorted(normalized.split(word_separator)))
         elif normalizer_option == 2:
             normalized = word_separator.join(sorted(set(normalized.split(word_separator))))
-        elif len(f_map) > 0:
+        elif len(f_map) > 0 and normalizer_option == 0:
             f_map[-1] = total_length - 1
             self.normalizer_result['map'] = f_map
+            self.normalizer_result['r_map'] = self.reverse_map(self.normalizer_result['map'])
         self.normalizer_result['normalized'] = normalized
         return normalized
 
